use serde::{Deserialize, Serialize};
use std::rc::Rc;

use super::core::MajorTickChannel;
use super::event::Event;
use super::{order, BetweenPriority, Service};
use crate::io::Packet;

use crate as fleck;
use crate::msg::MessageParams;

#[derive(Default)]
pub(crate) struct ParsePacket {}

impl ParsePacket {}

impl Service for Rc<ParsePacket> {
    fn register_channels(&self, eroot: &mut super::EventRoot) {
        // eroot.channel::<>().subscribe(self, Self::process_incoming);
    }
}

impl ParsePacket {
    fn process_incoming(&mut self, api: &crate::Fleck, _packet: &mut Packet) {
        // try deserializing
        // get message registry from api
        // remove data from packet, add parsed Msg instead
        println!("incoming packet!");
    }
}

pub struct SendPacketTag {}
pub type SendPacketChannel = (SendPacketTag, crate::io::Packet);

#[derive(Default)]
pub(crate) struct SendPacket {}

impl Service for Rc<SendPacket> {
    fn register_channels(&self, eroot: &mut super::EventRoot) {
        eroot.create_priority_channel::<SendPacketChannel>();
        eroot
            .priority_channel::<SendPacketChannel>()
            .with::<order::Last>()
            .subscribe(self, |s, api, packet| s.process_outgoing(api, packet));
    }
}

impl SendPacket {
    fn process_outgoing(&self, api: &crate::Fleck, packet: &mut Event<crate::io::Packet>) {
        // use default channel if not specified
        if packet.data.io_channel.is_none() {
            packet.data.io_channel = Some(api.raw_io().global().clone());
        }

        // serialize if needed
        if let (false, Some(msg)) = (packet.data.data.as_ref().is_some(), packet.data.msg.take()) {
            packet.data.data = Some(bincode::serialize(&msg).expect("failed to serialize message"));
        }

        match &mut packet.data.io_channel {
            Some(ch) => ch.clone().send_packet(&mut packet.data),
            _ => {
                println!("tried to send packet with no channel?");
            }
        }
    }
}

#[derive(Default)]
pub(crate) struct LocalDiscovery;

#[derive(Serialize, Deserialize)]
struct DiscoveryMsg {
    pkey: crate::crypto::PublicKey,
}

impl MessageParams for DiscoveryMsg {
    const NAME: &'static str = "DiscoveryMsg";
    const ENCRYPTED: bool = false;
}

impl Service for Rc<LocalDiscovery> {
    fn register_msgs(&self, registry: &mut fleck::msg::MessageRegistry) {
        registry.add_message_type::<DiscoveryMsg>();
    }

    fn register_channels(self: &Self, eroot: &mut super::EventRoot) {
        log::info!("LocalDiscovery registering channels...");
        eroot
            .channel::<MajorTickChannel>()
            .subscribe(self, LocalDiscovery::process_major_tick);
    }
}

impl LocalDiscovery {
    fn process_major_tick(&self, api: &crate::Fleck, _: &mut Event<()>) {
        log::info!("LocalDiscovery major tick!");

        api.queue_priority::<SendPacketChannel>(Packet {
            addr: Some((*crate::io::MULTICAST_ADDRESS).into()),
            data: None,
            io_channel: Some(api.raw_io().local().clone()),
            msg: Some(crate::msg::Message::build(DiscoveryMsg {
                pkey: Default::default(),
            })),
        });
    }
}
